{% extends "layout.html" %}
{% set title = 'SphinxReport' %}

{% block body %}

<i>CGATReport</i> is a report generator that is implemented as an extension
to <a href="http://sphinx.pocoo.org/">Sphinx</a>.

<p> Its purpose is to facilitate writing scientific reports
interpreting large and changing datasets. It is designed to assist the
iterative analysis during the development of a computational
scientific pipeline as understanding of the data grows.  Once the
pipeline settles down, <i>SphinxReport</i> permits an easy transition
towards the automatic report generation needed when the pipeline is
run on new data sets.  <p>

<i>CGATReport</i> is easy to use and powerful enough to give all the
flexibility needed during the development of computational pipelines
and robustness during the production use of a pipeline.  It is
intended for developers and computational scientists with <a
href="http://www.python.org">python</a> scripting experience.

<p> <i>CGATReport</i> comes with all the batteries included thanks to
<a href="http://www.python.org">python</a> and <a
href="http://sphinx.pocoo.org/">Sphinx</a>, in particular, it

<ul>
<li>uses simple markup in the form of <a href="http://docutils.sourceforge.net/rst.html">restructured text</a>,
<li>supports both automated and narrative analysis,
<li>keeps code, data and text together,
<li>takes care of indexing and formatting to produce something presentable,
<li>produces static html, latex, ps and pdf.
</ul>

<h1>How to get CGATReport</h1>

To install CGATReport, simply type:

<pre>
    pip install CGATReport
</pre>

The development version of CGATReport is available on
<A HREF="https://github.com/AndreasHeger/CGATReport">Github</A>.

<h1> Documentation</h1>

<table class="contentstable" align="center" style="margin-left: 30px"><tr>
  <td width="50%">
    <p class="biglink"><a class="biglink" href="contents.html">Contents</a><br/>
    <span class="linkdescr">for a complete overview</span></p>
    <p class="biglink"><a class="biglink" href="search.html">Search page</a><br/>
    <span class="linkdescr">search the documentation</span></p>
  </td><td width="50%">
    
    <p class="biglink"><a class="biglink" href="genindex.html">General Index</a><br/>
    <span class="linkdescr">all functions, classes, terms</span></p>
    <p class="biglink"><a class="biglink" href="modindex.html">Module Index</a><br/>
    <span class="linkdescr">quick access to all documented modules</span></p>
  </td></tr>
</table>

<h2>Examples</h2>

<B><A HREF="../examples/rnaseq/contents.html">RNA-Seq report</A></B> - 
a report generated from the <A HREF='http://www.cgat.org'>CGAT</A>
<A HREF="http://en.wikipedia.org/wiki/RNA-Seq">RNA-Seq</A> analysis pipeline using the data from
<A HREF="http://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE24399">Yang et. al. 2011</A>.


{% endblock %}
